#! /usr/bin/python3

import ast
import sys
import os
import re
import subprocess


def is_import_stmt(node):
    return type(node) == ast.Call and type(node.func) == ast.Name\
        and node.func.id == '__import__' and node.args\
        and type(node.args[0]) == ast.Constant


def is_importlib_call(node):
    return type(node) == ast.Call and type(node.func) == ast.Attribute\
            and type(node.func.value) == ast.Name\
            and node.func.value.id == 'importlib'\
            and node.func.attr == 'import_module'\
            and type(node.args[0]) == ast.Constant


def requ_check(code, path):  # For REQ in ['slight', 'relaxed']
    dep = set()
    rel_dep = set()
    advanced_dep = set()
    for node in ast.parse(code).body:
        if type(node) == ast.Import:
            for j in node.names:
                dep.add(j.name)
        elif type(node) == ast.ImportFrom:
            if node.level != 0:
                if behavior != 'original':
                    rel_dep.add(parents_package(path, node.level, node.module))
            else:
                dep.add(node.module)
        elif 'value' in dir(node) and is_import_stmt(node.value):
            try:
                if type(node.value.args[0]) == ast.BinOp\
                        and type(node.value.args[0].left.value) != ast.Attribute:
                    advanced_dep.add(node.value.args[0].left.value)
                elif type(node.value.args[0].value) != ast.Name and type(node.args[0].value) == str:
                    advanced_dep.add(node.value.args[0].value)
            except Exception as msg:
                print(msg, file=sys.stderr)
                pass

    return dep, rel_dep, advanced_dep


def parents_package(path, level=0, dep_name=None, module=''):
    if not level:
        level = - len(path)

    if not module:
        for suffix in reversed(sorted(sys.path, key=lambda pth: pth.count('/'))):
            if suffix and (m := re.findall(r'\S*%s/(\w*)[/\.]' % suffix,  path)):
                module = m[0]
                break

    return '.'.join([p for p in re.search(r'%s(\.py|/\S*)' % module, path).group().split('/')[:- level] + ([dep_name] if dep_name else []) if p])


def read_ast_tree(codes, path, REQ):
    dep = set()
    relative_dep = set()
    advanced_dep = set()
    try:
        for node in ast.walk(ast.parse(codes)):
            if type(node) == ast.Import:
                for j in node.names:
                    dep.add(j.name)
            elif type(node) == ast.ImportFrom:
                if node.level != 0:
                    relative_dep.add(parents_package(path,
                                     node.level, node.module))
                else:
                    dep.add(node.module)
            elif is_import_stmt(node):
                try:
                    if type(node.args[0].value) != ast.Name\
                            and type(node.args[0].value) == str:
                        advanced_dep.add(node.args[0].value)
                    else:
                        raise
                except Exception as msg:
                    print(msg, file=sys.stderr)
                    if sys.argv.count('--verbose') != 0:
                        print('python3.req.py:%s: failed to determine the name\
                               of the module contained from expression: %s'
                              % (path, node.unparse), file=sys.stderr)
            elif is_importlib_call(node):
                advanced_dep.add(node.args[0].value)
    except SyntaxError:
        print('python3.req: error: %s: invalid syntax'
              % (path), file=sys.stderr)
        if file_name[-3:] == '.py' or len(code) > 0\
                and code[0].startswith('#!'):
            pass
        else:
            print('python3.req: %s: possibly not pythonish file'
                  % file_name, file=sys.stderr)
    return dep, relative_dep, advanced_dep


def appender(prov_material, module):
    prov = list()

    index = prov_material.index(module)

    prov.append(prov_material[0])

    for sub in prov_material[1: index + 1]:
        prov.append('%s.%s' % (sub, prov[-1]))
    return prov


def scanner(path, module=None):
    prov = list()

    try:
        if os.path.isfile(path):
            if path.endswith('.py'):
                path, trash = os.path.splitext(path)
            elif os.path.isfile(path) and path.endswith('.so'):
                path = path.split('/')
                path[-1] = path[-1].split('.')[0]
                path = '/'.join(path)
            else:
                return prov

            if module:
                prov.append(parents_package(path=path, module=module))
                if path.endswith('__init__'):
                    prov.append(parents_package(path=path.replace('__init__', ''), module=module))
            else:
                pm = [p for p in path.split('/') if p]
                pm.reverse()
                module = pm[-1]

                prov = appender(pm, module)
                if path.endswith('__init__'):
                    prov += appender(pm[1:], module)

        elif os.path.isdir(path):
            for submodule in os.scandir(path):
                if submodule.name != '__pycache__':
                    prov += scanner(submodule.path, module)

    except PermissionError:
        print('Warning! Cannot inspect %s, permission denied' % path, file=sys.stderr)

    return prov


def provides_creator(file_path):
    suffix = None

    for suffix in reversed(sorted(sys.path, key=lambda pth: pth.count('/'))):
        if suffix and (m := re.findall(r'\S*%s/(\w*)[/\.]' % suffix,  file_path)):
            module = m[0]
            file_path = re.match(r'\S*%s/\w*' % suffix, file_path).group()
            return scanner(file_path, module), module
    return scanner(file_path), None


def catch_so(name):
    dep_version = os.getenv('RPM_PYTHON3_VERSION', "%s.%s" % sys.version_info[0:2])

    res = subprocess.run(['objdump', '-p', name], capture_output=True, stderr=None)
    if res.returncode != 0:
        print(res.stderr.decode(), file=sys.stderr)
        return
    if res.stdout.find(b'file format elf64') != -1:
        dep_suffix = "(64bit)"
    else:
        dep_suffix = ""
    print("python%s-ABI%s" % (dep_version, dep_suffix))


def ignore(name):
    if IS_HIER is None:
        name = name.split('.')[0]
    if name in dict([(x, 1) for x in list(os.getenv('RPM_PYTHON3_REQ_SKIP', "").split()) + list(sys.builtin_module_names)]):
        return True
    elif name == '__main__':
        return True
    else:
        return False


def get_raw_encoding(line):
    enc = re_coding.search(line)
    if enc:
        return enc.group(1)


# Parser/tokenizer.c:get_normal_name()
def get_enc_normal_name(enc):
    enc = enc.lower().replace('_', '-')
    if enc.startswith('utf-8'):
        return 'utf-8'
    for alias in ['latin-1', 'iso-8859-1', 'iso-latin-1']:
        if enc.startswith(alias):
            return 'iso-8859-1'
    return enc


missing_encodings = None


def get_encoding_module(enc, file_name):
    mod = 'encodings'
    try:
        from encodings import search_function
    except Exception as msg:
        print(msg, file=sys.stderr)
        global missing_encodings
        missing_encodings = True
        return mod
    try:
        mod = search_function(enc)[1].__self__.__class__.__module__
        print("python3.req: %s: encoding=%s module=%s" % (file_name, enc, mod), file=sys.stderr)
    except Exception as msg:
        print(msg, file=sys.stderr)
        pass
    return mod


def need_encoding_module(line, file_name):
    enc = get_raw_encoding(line)
    if enc:
        enc = get_enc_normal_name(enc)
        if enc not in ['utf-8', 'iso-8859-1']:
            print("python3.req: %s: non-standard encoding: %s" % (file_name, enc), file=sys.stderr)
            mod = get_encoding_module(enc, file_name)
            return mod


def printing(dependencies, filters, relative_flag, file_name, REQ, module=None):
    head, tail = os.path.split(file_name)
    for dep in dependencies:
        if ignore(dep):  # Исключаются те модули, которые либо уже предоставлены конкретной данной сборкой python3, либо исключены средой сборки 
            print("python3.req: %s: skipping %s"
                  % (file_name, dep), file=sys.stderr)

        elif dep in adj_provides or dep in provides and any([not module, module and relative_flag == True]):
            print(f'{file_name}: {dep} is possibly a self-providing dependency, skip it', file=sys.stderr)

        elif not (dep in filters) and REQ in ['slight', 'relaxed']:
            print('%s: %s: Ignore for REQ=%s module=%s' % (sys.argv[0], file_name, REQ, dep), file=sys.stderr)

        elif relative_flag == False:  # Если зависимости были получены import
            if IS_HIER is None:
                dep = str(dep).split('.')[0]
            print("%s(%s)" % ("python%u" % sys.version_info[0], dep), end='\n')

        elif relative_flag:  # Если зависимсти были получены __import__ или from import
            print("%s(%s)" % ("python%u" % sys.version_info[0], dep), end='\n')

        else:
            print(f'{sys.argv[0]}: Do not know what to with dep:{dep}',
                  end='\n', file=sys.stderr)


behavior = os.getenv('RPM_PYTHON3_BEHAVIORS', '')
if behavior == '':
    behavior = 'original'

REQ = os.getenv('RPM_PYTHON3_REQ_METHOD', 'strict')
IS_HIER = os.getenv('RPM_PYTHON3_REQ_HIER', None)
if IS_HIER and IS_HIER.startswith('no') or sys.argv.count('--no-submodules') != 0:
    # rpm-build-python/tools/requires.py uses 'hier' as the default value;
    # rpm-build-python3/python3 can set it to 'yes'.
    # So, let's suppose the opposite value can be either 'nohier' or 'no'
    # (or simply an unset env variable). Here we map those possibilities
    # to our internal canonical representation of this value
    # (just in case someone may use them).
    IS_HIER = None

file_list = []
if len(sys.argv) > 1:
    for x in sys.argv[1:]:
        if x[:2] != '--':
            file_list.append(x)
else:
    file_list = sys.stdin.read().split()

re_coding = re.compile(r'^\s*#.*?coding[:=]\s*([\w._-]+)')
provides = set()
modules = set()

for file_name in file_list:
    tmp_prov, module = provides_creator(file_name)
    provides = provides.union(set(tmp_prov))

    if module and re.match(r'\S*%s\.(py|[^/]*so)' % module, file_name):
        provides.add(module)

new_provides = os.getenv('RPM_PYTHON3_SELF_PROV_PATH')
adj_provides = set()
if new_provides is not None:
    for path in new_provides.split(':'):
        adj_provides = adj_provides.union(set(scanner(path)))

for file_name in file_list:
    if str(file_name)[-3:] == '.so':
        catch_so(file_name)
    elif not str(file_name)[-4:] in ['.pyc', '.pyo', '.pth']:
        try:
            f = open(file_name, encoding="utf-8-sig")
        except PermissionError:
            continue
        except FileNotFoundError:
            continue
        except IsADirectoryError:
            continue

        try:
            lines = []
            for x in f:
                lines.append(x)
        except Exception as msg:
            print(msg, file=sys.stderr)
            lines = []
        f.close()

        lines.append('')
        code = ''.join(lines)
        dependencies, relative_dependencies, advanced_dependencies = read_ast_tree(code, file_name, REQ)
        if sys.argv.count('--no-encodings') == 0:
            encmod = None
            if len(lines) > 0:
                encmod = need_encoding_module(lines[0], file_name)
            if not encmod and len(lines) > 1:
                encmod = need_encoding_module(lines[1], file_name)
            if encmod:
                print("%s(%s)" % ("python%u" % sys.version_info[0], encmod), end='\n')

        filtered_adv_dep = []
        filtered_rel_dep = []
        filtered_dep = []
        if REQ in ['slight', 'relaxed']:
            filtered_dep, filtered_rel_dep, filtered_adv_dep = requ_check(
                    code, file_name)

        dependencies = sorted(dependencies)
        relative_dependencies = sorted(relative_dependencies)
        advanced_dependencies = sorted(advanced_dependencies)

        printing(dependencies, filtered_dep, False, file_name, REQ)
        printing(relative_dependencies, filtered_rel_dep,
                 True, file_name, REQ, module)
        printing(advanced_dependencies, filtered_adv_dep,
                 'Adv', file_name, REQ)
